def main():
    x = Ax()
    y = Ax()
    z = Ax()
    w = Ax()
    # (x -o y) -o (z -o w) -o x*z -o y*w
    a = Tensor(x.neg, y.neg)
    b = Tensor(z.neg, w.neg)
    c = Par(x.pos, z.pos)
    d = Tensor(y.pos, w.pos)
    cd = Par(c, d)
    bcd = Par(b, cd)
    abcd = Par(a, bcd)
    p4 = Bq(z, Bq(Bq(x, y, a), w, d), b)
    p0 = Uq(Uq(Uq(Uq(p4, c), cd), bcd), abcd)
    conj_map = Root(p0)
    # We should run few cut elimination scenarios to see if they run correctly.
    # Trigger left variable substitution
    x = Ax()
    y = Ax()
    xy = Par(x.pos, y.neg)
    cut = Cut(x.neg, y.pos)
    p0 = Uq(Bq(x, y, cut), xy)
    root = Root(p0)
    print list(cut_elimination(cut))
    full_check(root.body, root)
    # Trigger tensor-par match
    x = Ax()
    z = Ax()
    w = Ax()
    xw = Par(x.neg, w.neg)

    l = Tensor(x.pos, w.pos)
    r = Par(z.neg, z.pos)
    pl = Bq(x, w, l)
    pr = Uq(z, r)
    cut = Cut(l, r)
    p0 = Uq(Bq(pl,pr, cut), xw)
    root = Root(p0)
    x, y = list(cut_elimination(cut))
    full_check(root.body, root)
    print list(cut_elimination(x))
    full_check(root.body, root)
    print list(cut_elimination(y))
    full_check(root.body, root)
    # Trigger left switch

# The following structures encode the different steps in proofs.
class Ax(object):
    def __init__(self):
        self.parent = None
        self.neg = provide(Line(), self)
        self.pos = provide(Line(), self)

class Uq(object):
    def __init__(self, body, op):
        self.parent = None
        self.body = child(body, self)
        self.op = provide(op, self)

class Bq(object):
    def __init__(self, left, right, op):
        self.parent = None
        self.left = child(left, self)
        self.right = child(right, self)
        self.op = provide(op, self)

class Cq(object):
    def __init__(self, left, right):
        self.parent = None
        self.left = child(left, self)
        self.right = child(right, self)

# To avoid difficult questions with cut elimination there's a root object.
class Root(object):
    def __init__(self, body):
        self.body = child(body, self)
        full_check(self.body, self)

# The connectives are represented with graph structure that is traversed
# along the sequent proofs.
class Line(object):
    def __init__(self):
        self.top = None
        self.bot = None

class Cut(object):
    def __init__(self, left, right):
        self.left = plug(left, self)
        self.right = plug(right, self)
        self.top = None

class Tensor(object):
    def __init__(self, left, right):
        self.left = plug(left, self)
        self.right = plug(right, self)
        self.top = None
        self.bot = None

class Par(object):
    def __init__(self, left, right):
        self.left = plug(left, self)
        self.right = plug(right, self)
        self.top = None
        self.bot = None

# To verify we got a well-formed proof, the slots and fields are checked.
def full_check(seq, parent):
    assert seq.parent is parent
    if isinstance(seq, Ax):
        assert seq.pos.top is seq
        assert seq.pos.bot is not None
        assert seq.neg.top is seq
        assert seq.neg.bot is not None
    elif isinstance(seq, Uq):
        full_check(seq.body, seq)
        check_uq(seq, seq.op)
    elif isinstance(seq, Bq):
        full_check(seq.left, seq)
        full_check(seq.right, seq)
        check_bq(seq, seq.op)
    else:
        assert False

# The flow-graph in the connectives is ensured to satisfy the scoping rules.
def check_uq(seq, op):
    if isinstance(op, Par):
        assert op.left.bot is op
        assert op.right.bot is op
        assert in_scope(op.left.top, seq)
        assert in_scope(op.right.top, seq)
        assert op.top is seq
        assert (op.bot is not None) or isinstance(op.top.parent, Root)
    else:
        assert False

def check_bq(seq, op):
    if isinstance(op, Cut):
        assert op.left.bot is op
        assert op.right.bot is op
        assert in_scope(op.left.top, seq.left)
        assert in_scope(op.right.top, seq.right)
        assert op.top is seq
    elif isinstance(op, Tensor):
        assert op.left.bot is op
        assert op.right.bot is op
        assert in_scope(op.left.top, seq.left)
        assert in_scope(op.right.top, seq.right)
        assert op.top is seq
        assert (op.bot is not None) or isinstance(op.top.parent, Root)
    else:
        assert False

# Most manipulations to the structure are difficult due to interconnections,
# so there are many small utility functions to ensure the code here
# correctly manipulates the structure.
def child(obj, this):
    assert obj.parent is None
    obj.parent = this
    return obj

def unchild(obj, this):
    assert obj.parent is this
    obj.parent = None
    return obj

def provide(obj, this):
    assert obj.top is None
    obj.top = this
    return obj

def plug(obj, this):
    assert obj.bot is None
    obj.bot = this
    return obj

def in_scope(obj, scope, stop=None):
    walk = obj
    while walk is not scope:
        walk = walk.parent
        if walk is stop:
            return False
    return True

# The C-reduction eliminates cuts.
# Note the reduction always happens to the "left" side first.
# Proof-equivalence in linear logic is more challenging thing.
def cut_elimination(cut):
    top = cut.top
    if isinstance(cut.left, Line):
        pivot_in(top, top.left, top.right, cut.left, cut.right)
    elif isinstance(cut.right, Line):
        pivot_in(top, top.right, top.left, cut.right, cut.left)
    elif top.left is not cut.left.top:
        replace(top, unchild(top.left, top))
        replace(cut.left.top, top)
        top.left = child(cut.left.top, top)
    elif top.right is not cut.right.top:
        replace(top, unchild(top.right, top))
        replace(cut.right.top, top)
        top.right = child(cut.right.top, top)
    elif isinstance(cut.left, Tensor) and isinstance(cut.right, Par):
        return cut_tensor(top, cut.right, cut.left)
    elif isinstance(cut.left, Par) and isinstance(cut.right, Tensor):
        return cut_tensor(top, cut.left, cut.right)
    else:
        assert False
    return iter(())

def pivot_in(top, chain, box, pivot, val):
    ax = pivot.top
    replace(top, unchild(chain, top))
    replace(ax, unchild(box, top))
    dual = {ax.neg:ax.pos, ax.pos:ax.neg}[pivot]
    if dual.bot.left is dual:
        dual.bot.left = val
        val.bot = dual.bot
    elif dual.bot.right is dual:
        dual.bot.right = val
        val.bot = dual.bot
    else:
        assert False

def cut_tensor(top, par, tensor):
    body = unchild(par.top.body, par.top)
    left = unchild(tensor.top.left, tensor.top)
    right = unchild(tensor.top.right, tensor.top)
    c0 = Cut(detach(tensor.left), detach(par.left))
    c1 = Cut(detach(par.right), detach(tensor.right))
    body = Bq(left, body, c0)
    body = Bq(body, right, c1)
    replace(top, body)
    yield c0
    yield c1

# Utilities to implement the 'cut'
def detach(this):
    this.bot = None
    return this

def replace(this, that):
    if isinstance(this.parent, Bq):
        if this.parent.left is this:
            this.parent.left = child(that, this.parent)
        elif this.parent.right is this:
            this.parent.right = child(that, this.parent)
        else:
            assert False
    elif isinstance(this.parent, Uq):
        if this.parent.body is this:
            this.parent.body = child(that, this.parent)
        else:
            assert False
    else:
        assert False
    this.parent = None

if __name__=='__main__':
    main()
